@@ -1,6 +1,6 @@
 /*
   Simple DirectMedia Layer
-  Copyright (C) 1997-2013 Sam Lantinga <slouken@libsdl.org>
+  Copyright (C) 1997-2021 Sam Lantinga <slouken@libsdl.org>
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -18,6 +18,9 @@
      misrepresented as being the original software.
   3. This notice may not be removed or altered from any source distribution.
 */
+#include "../../SDL_internal.h"
+
+#ifdef SDL_TIMERS_PSP
 
 #include "SDL_thread.h"
 #include "SDL_timer.h"
@@ -31,7 +34,8 @@
 static struct timeval start;
 static SDL_bool ticks_started = SDL_FALSE;
 
-void SDL_InitTicks(void)
+void
+SDL_TicksInit(void)
 {
     if (ticks_started) {
         return;
@@ -41,24 +45,29 @@ void SDL_InitTicks(void)
     gettimeofday(&start, NULL);
 }
 
-Uint32 SDL_GetTicks(void)
+void
+SDL_TicksQuit(void)
 {
+    ticks_started = SDL_FALSE;
+}
+
+Uint64
+SDL_GetTicks64(void)
+{
+    struct timeval now;
+
     if (!ticks_started) {
-        SDL_InitTicks();
+        SDL_TicksInit();
     }
 
-    struct timeval now;
-    Uint32 ticks;
-
     gettimeofday(&now, NULL);
-    ticks=(now.tv_sec-start.tv_sec)*1000+(now.tv_usec-start.tv_usec)/1000;
-    return(ticks);
+    return (Uint64)(((Sint64)(now.tv_sec - start.tv_sec) * 1000) + ((now.tv_usec - start.tv_usec) / 1000));
 }
 
 Uint64
 SDL_GetPerformanceCounter(void)
 {
-    return SDL_GetTicks();
+    return SDL_GetTicks64();
 }
 
 Uint64
@@ -75,5 +84,7 @@ void SDL_Delay(Uint32 ms)
     sceKernelDelayThreadCB(ms * 1000);
 }
 
+#endif /* SDL_TIMERS_PSP */
+
 /* vim: ts=4 sw=4
  */
